#!perl
# CVS: $Id: ljsm.pl,v 2.5 2003/12/21 20:15:16 sasha Exp $
# Author: Alexander Nikolaev <sasha_nikolaev@yahoo.com>
# Descr: fetch all user's posts and/or memories and save them in HTML format

=head1 SYNOPSYS

see usage() subroutine for usage summary

=head1 SETUP

 I've tested this script with Windows Me, ActiveState perl v. 5.6.0
 with the following ppm modules installed:
 HTML-Parser 2.33, HTML-SimpleLinkExtor 0.71, libwww-perl 5.48
 If you want to use UTF-8 conversion to local charset you'll also 
 need to setup Unicode-MapUTF8. 

 -------- How to setup Unicode-MapUTF8 for Windows ------------
 1. ppm install Unicode-Map
 2. download, untar and ungzip Unicode-MapUTF8
 3. install Unicode-MapUTF8 ignoring messages about missing modules
 4. go to %PERL%\site\lib\Unicode and edit MapUTF8.pm:
 	4.1 comment out 'use Unicode::Map8' and 'use Jcode'
 	4.2. insert 'use Unicode::Map' line if it is not already there
 --------------------------------------------------------------

=head1 TODO

 - [?] fix -d and -O interaction
 - [x] add charset pragma to index.html
 - [x] add command-line switch for image download
 - [x] add &usescheme=lynx to lj queries.
 - [x] proxy support
 - [-] usable pager
 - [x] better date range handling (command-line switch?), 
 - [x] explicitly show 
 	"You must be logged in to view this protected entry" and 
	"This journal is deleted." cases in the index file
 - [x] XML export
 - [x] no comments download

=head1 LINKS

http://www.offtopia.net/~ati/ljsm/
http://www.livejournal.com/talkread.bml?journal=ru_hitech&itemid=118529
http://www.livejournal.com/talkread.bml?itemid=122758&journal=ru_hitech
http://www.livejournal.com/talkread.bml?journal=ru_hitech&itemid=158872
http://www.livejournal.com/talkread.bml?itemid=394253&journal=rulj
http://www.livejournal.com/community/rulj/532637.html
http://www.livejournal.com/community/rulj/595146.html

=head1 SUBROUTINES

=cut

use constant LOGIN			=> '';	# leave it empty if you don't want to login
use constant PASSWORD		=> '';
use constant START_YEAR		=> 2001;	# fetch data back to this year
use constant UTF8_DECODE	=> 0;		# convert text to local charset
use constant LOCAL_CHARSET	=> 'CP1251'; # windows cyrillic
use constant DEBUG_LEVEL	=> 3;		# 0 - quiet, 1 - essential, 2 - verbose
use constant LOCAL_DIR		=> '';		# local directory to put files into. Leave it empty to put in the current directory. Slash (/, if not empty) in the end required.
use constant HTTP_PROXY		=> '';		# set proxy URL if you use http proxy
use constant CVSVERSION		=> '$Revision: 2.5 $'; # don't touch this
use constant SAVE_PICS		=> 1;	# download standard icons (1) usepics (2) or all graphics referenced by post (3)


# ===================================================================
# end of public constants definition. no user-editable parts below this line
# ===================================================================
use constant MAX_TRIES		=> 5; # max tries to get page in case of failure
use constant BASE_URL		=> 'http://www.livejournal.com/';
use constant CATALOG_URL	=> BASE_URL . 'view/?type=month';
use constant LOGIN_SCRIPT	=> 'login.bml';
use constant POST_SCRIPT	=> 'talkread.bml';
use constant POST_SCRIPTNOC	=> 'talkpost.bml';
use constant MEMO_SCRIPT	=> 'tools/memories.bml';
use constant EXPORT_SCRIPT	=> 'export.bml';

use LWP::UserAgent;
use HTTP::Cookies;
use HTML::Form;
use HTML::SimpleLinkExtor;
use File::Path;
use File::Basename;
use File::Find;
if (UTF8_DECODE) {
	use Unicode::MapUTF8 qw(from_utf8);
}
use Compress::Zlib;
use Getopt::Std;
#use Data::Dumper;
use strict;

my ($ua, $req, $res, $login, @posts, %images, $user, %stat, %memories, %posts, $umask);
our ($opt_r, $opt_m, $opt_a, $opt_c, $opt_O, $opt_i, $opt_I, $opt_u, $opt_U, $opt_x, $opt_X, $opt_p, $opt_d);

# open log file (delete it if there were no errors)
$umask = umask 0077;
open LF, ">>ljsm.log" or die "error opening ljsm.log for appending: $!\n";
umask $umask;
print LF "\n============= " . join(' ', @ARGV) . "\n";
print LF scalar localtime() . "\n";

# steal options from @ARGV before we go for users
getopts('rmacxXOIUu:p:d:i:');
$opt_i = SAVE_PICS if (!$opt_i);

usage() && exit unless (@ARGV || ($opt_u && $opt_X));


# rebuild indexes and exit if -x option is set
if ($opt_x) {
	foreach $user (@ARGV) { # for each user
		logmsg("rebuilding index file for user $user...\n");
		build_index($user);
		logmsg("done.\n");
	}
	exit 0;
}

# init global vars
$stat{$_} = 0 foreach ('users','pages_ok','got_posts','images');
%images = ();

$ua = new LWP::UserAgent;
$ua->cookie_jar(new HTTP::Cookies(
	file			=> "ljcookies.txt",
	autosave		=> 0)
);

# set proxy URL for LWP requests
$ua->proxy('http', HTTP_PROXY) if HTTP_PROXY;
if ($opt_p) {
	$opt_p = "http://$opt_p" unless ($opt_p =~ m{^http://});
	$ua->proxy('http', $opt_p);
}

# get cookies
exit 1 unless (!(LOGIN || $opt_u) || ($login = lj_login()));

# get xml files and exit if -X option is set
if ($opt_X) {
	if (!$login) {
		logmsg("only registered user can export her posts as XML files\n");
		logmsg("please specify -u login:password at the command line\n");
		exit 128;
	}
	($user) = split(':', $opt_u);
	get_xml($user);
	exit 0;
}

# get posts and memories
foreach $user (@ARGV) { # for each user
	%memories = %posts = ();
	$stat{'count_posts'} = $stat{'count_memos'} = 0;
	logmsg("\n\n=== processing user $user\n");
	@posts = ();
	
	push @posts, get_memos($user) if ($opt_m || $opt_a);
	push @posts, get_posts($user) unless ($opt_m && !$opt_a);
	get_files($user);
	build_index($user) if @posts;
	undef @posts; # free memory
	$stat{'users'}++;
}

# get images
if (($stat{'got_posts'} > 0) && 
	(scalar keys %images) && 
	$opt_i) {
	get_pics();
}



#   ============================================
#                  subroutines
#   ============================================

=item get_date_range($user)

get year and month of the last downloaded post

=cut
sub get_date_range {
	my ($user, $is_xml) = @_;
	
	my ($start_year, $start_month, $end_year, $end_month, @date, $t);
	
	@date = localtime();

	# get end date
	if ($opt_d) {
		($start_year, $start_month, $end_year, $end_month) = split(/\D/, $opt_d);
		
		$end_year = $date[5]+1900 unless $end_year;
		$end_month = $date[4]+1 unless $end_month;
		
		# swap dates if specified in reversed order
		if ($start_year > $end_year) {
			($start_year, $end_year, $start_month, $end_month) = ($end_year, $start_year, $end_month, $start_month);
		} elsif (($start_year == $end_year) && ($start_month > $end_month)) {
			($start_month, $end_month) = ($end_month, $start_month);
		}
		
		return ($start_year, $start_month, $end_year, $end_month);
			
	} else {
		$start_year = START_YEAR;
		$start_month = 1;
		$end_year  = $date[5] + 1900;
		$end_month = $date[4] + 1;
	}
	
	# set start_year, start_month based on the downloaded posts
	if (!(-d LOCAL_DIR . $user) || $opt_O || $opt_r) { 
		return ($start_year, $start_month, $end_year, $end_month);
	}
	
	if (!$is_xml) { # date range between last post and current month
		opendir(UD, LOCAL_DIR . $user) or die "error opening " . LOCAL_DIR . "$user directory for reading: $!\n";
		my ($year) = sort {$b <=> $a } grep(/^\d+$/, readdir(UD));
		closedir UD;
		return ($start_year, $start_month, $end_year, $end_month) unless $year;
		
		opendir(UD, LOCAL_DIR . "$user/$year") or die "error opening " . LOCAL_DIR . "$user/$year directory for reading: $!\n";
		my ($month) = sort {$b <=> $a } grep(/^\d+$/, readdir(UD));
		closedir UD;
		$month = 1 unless $month;
		
		return ($year, $month, $end_year, $end_month);
		
		
	} else { # date range for XML export
		opendir(UD, LOCAL_DIR . $user . '/export') or die "error opening " . LOCAL_DIR . "$user/export directory for reading: $!\n";
		my ($lastfile) = reverse sort grep (/^\d+_\d+\.xml$/, readdir UD);
		closedir UD;
		if ($lastfile && ($lastfile =~ /^(\d+)_(\d+)/)) {
			$start_year = $1;
			$start_month = $2;
		}
		return ($start_year, $start_month, $end_year, $end_month);
	}
}




=item get_pics()

download userpics, buttons etc

=cut
sub get_pics {
	my ($imgsrc, $img);

	logmsg("getting pictures...\n",2);
	foreach $imgsrc (keys %images) {
		# test if there is already image with the same name
		next if (-f $images{$imgsrc});
		
		# get image
		if ($img = get_page($imgsrc, 1)) {
		 	mkpath(dirname($images{$imgsrc}), DEBUG_LEVEL, 0755) 
				unless -d dirname($images{$imgsrc});
			if (open (DF, ">$images{$imgsrc}")) {
				binmode DF;
				print DF $img;
				close DF;
				$stat{'images'}++;
			} else {
				logmsg("error opening $images{$imgsrc} for writing: $!\n",0);
			}
		} else {
			logmsg("error getting $imgsrc\n",0);
		}
	}
}



sub get_xml {
	my ($user) = @_;
	my ($path, $content, $tmp, $exportform, $month, $emonth, $year, $fname);
	
	logmsg("exporting $user\'s posts in XML format...\n",2);
	$path = LOCAL_DIR . $user . '/export';
	mkpath($path, DEBUG_LEVEL, 0755) if (!-d $path);
	my ($start_year, $start_month, $end_year, $end_month) = get_date_range($user, 1);

	# get export form
	if ($content = get_page(BASE_URL . EXPORT_SCRIPT)) {
		($tmp, $exportform) = parse HTML::Form($content, BASE_URL);
		if (defined $exportform) {
			#$exportform->dump();
			$exportform->value('format', 'xml');
			$exportform->value('encid', 3); # 6 for KOI-8
			# uncomment next line to skip translation between encodings
			# $exportform->value('notranslation', 1);
			
			$year = $end_year;
			YEAR:
			while ($year >= $start_year) {
				$emonth = ($year == $start_year)? $start_month : 1;
				for ($month = 12; $month >= $emonth; $month--) {
					next if (($year == $end_year) && ($month > $end_month));
					$fname = $path . sprintf("/%4d_%02d.xml", $year, $month);
					
					if (-f $fname) {
						if ($opt_O) {
							logmsg("!! overwriting $fname\n",2);
						} elsif ($opt_r) {
							logmsg("-r: skipping $fname\n", 2);
							next;
						} else {
							last YEAR;
						}
					}

					logmsg("<< $year/$month\n",2);
					$exportform->value('year', $year);
					$exportform->value('month', $month);
					
					#submit export form
					$req = $exportform->click();
					$req->header('Accept-Encoding' => 'gzip;q=1.0, *;q=0');
					
					foreach (0 .. MAX_TRIES) {
						$res = $ua->request($req);
						logmsg("retrying $year/$month...\n") if $_;
						if ($res->is_success) {
							$content = ($res->content_encoding && ($res->content_encoding =~ /gzip/))?
								Compress::Zlib::memGunzip($res->content) : $res->content;
							# save contents to a file
							logmsg(">> $fname\n");
							open (DF, ">$fname") or die "error opening $fname for writing: $!\n";
							print DF $content;
							close DF;
							last;
							
						} else {
							# TODO add logging of failed xml exort
							my $err = $res->error_as_HTML;
							$err =~ s/^[^\d].*$//mg;
							$err =~ s/[\n\r]+//g;
							logmsg("\n$err\n",0);
							if ($_ == MAX_TRIES) {
								logmsg ("failed to get XML file for $year/$month\n");
								print LF "Failed: xml for $year/$month\n";
								$stat{'pages_err'}++;
								return 0 unless $opt_I;
							}
							sleep 2;
						}
					}
				}
				$year--;
			}
			
		} else {
			logmsg("error getting export form " . BASE_URL . EXPORT_SCRIPT);
			return 0;
		}
	}
}


=item get_memos($user)

get list of user's memories and store them is $posts{memos}

=cut
sub get_memos {
	my ($user) = @_;
	my($extor_kw, $extor_posts, $content, $amuser, $keyword);
	my (@memos, $link, $link_post);
	
	logmsg("getting list of memories...\n",2);
	# get list of keywords
	if ($content = get_page(BASE_URL . MEMO_SCRIPT . "?user=$user")) {
		$extor_kw = new HTML::SimpleLinkExtor(BASE_URL);
		$extor_kw->parse($content);
		
		foreach $link ($extor_kw->a) {
			# get list of keywords links
			next unless $link =~ /@{[MEMO_SCRIPT]}\?user=\w+\&keyword=([^\&]+)\&filter=all$/;
			$keyword = $1;
			$keyword = " " unless length $keyword;
			# unescape keywords
			$keyword =~  s/\+/ /g;
			$keyword =~  s/%([0-9A-Fa-f]{2})/chr(hex($1))/eg;
			$keyword = from_utf8({-string => $keyword, -charset => LOCAL_CHARSET}) if (UTF8_DECODE || $opt_U);
			
			# get list of posts for the given keyword
			if ($content = get_page($link)) { 
				$extor_posts = new HTML::SimpleLinkExtor(BASE_URL);
				$extor_posts->parse($content);
				
				foreach $link_post ($extor_posts->a) {
					next unless (
						($link_post =~ m#^@{[BASE_URL]}(community/|users/|~)(\w+)/(\d+)\.html#) ||
						($link_post =~ /@{[POST_SCRIPT]}\?(journal=(\w+)\&)?itemid=(\d+)/)
					);
					next unless ($opt_O || ! -f LOCAL_DIR . "$user/memories/$2\_$3.html");
					
					push @memos, {
						'type'		=> 'memo',
						'status'	=> 0,
						'amuser'	=> $2,
						'keyword'	=> $keyword,
						'link'		=> $link_post->as_string,
						'itemid'	=> $3 };
				}

			} else { # error fetching list of posts
				logmsg("error fetching list of posts for user $user, keyword $keyword",0);
			}
		}
	} else { # error fetching list of keywords
		logmsg("error fetching list of keywords for $user\n",0);
	}
	
	return @memos;
}


=item get_posts($user)

get list of user's posts and store them in $posts{posts}

=cut
sub get_posts {
	my ($user) = @_;
	my ($content, $year, $month, @posts, $extor, $link, $emonth);
	
	@posts = ();
	my ($start_year, $start_month, $end_year, $end_month) = get_date_range($user);
	logmsg("getting posts links for $user " .
		sprintf("[ %4d/%02d - %4d/%02d ]", $start_year,$start_month,$end_year,$end_month) . "\n");
	
	$year = $end_year;
	YEAR:
	while ($year >= $start_year) {
		$emonth = ($year == $start_year)? $start_month : 1;
		for ($month = 12; $month >= $emonth; $month--) {
			next if (($year == $end_year) && ($month > $end_month));

			#fetch catalog data
			if ($content = get_page(CATALOG_URL . "&user=$user&y=$year&m=$month")) {
				$extor = new HTML::SimpleLinkExtor(BASE_URL);
				$extor->parse($content);
				
				# process links.
				foreach $link (reverse sort $extor->a) {
					if ($link =~ m#/$user/(\d+)\.html#) {
						push @posts, {
							'type'		=> 'post',
							'status'	=> 0,
							'year'		=> $year,
							'month'		=> $month,
							'link'		=> $link->as_string,
							'itemid'	=> $1,
							'keyword'	=> ''
						};
					}
				} # link loop on the catalog page
				
			} else { # error fetching catalog data
				logmsg("!! Error fetching catalog data. Going on with " . @posts . " posts\n");
				return @posts unless $opt_I;
			}
		} # months loop
		$year--;
	} # years loop
	return @posts;
}


=item get_files($user)

download and process posts and memories.

=cut
sub get_files {
	my ($user) = @_;
	my ($post, $content, $dir, $fname, $result, $extor, $up, $navbar, $n, $myhref);
	
	logmsg("getting posts...\n");
	
	my $charset = "<meta http-equiv=\"Content-Type\" content=\"text/html; ";
	$charset .= (UTF8_DECODE || $opt_U)?  "charset=@{[LOCAL_CHARSET]}\">" : "charset=utf-8\">";

	foreach $post (@posts) {
		if ($post->{'type'} eq 'post') {
			$dir = LOCAL_DIR . "$user/$post->{year}/$post->{month}";
			$fname = "$post->{itemid}.html";
			$up = "../../..";
			
		} else { # memo
			$dir = LOCAL_DIR . "$user/memories";
			$fname = "$post->{amuser}_$post->{itemid}.html";
			$up = "../..";
		}
		
		if (-s "$dir/$fname") {
			if ($opt_O) {
				logmsg("!! overwriting $dir/$fname\n", 2);
			} elsif ($opt_r) {
				logmsg("-r: skipping $dir/$fname\n", 2);
				next;
			} else {
				last;
			}
		}
		
		# old scheme
		if ($post->{'link'} =~ m#(journal=\w+&itemid=\d+)#) {
			$myhref = POST_SCRIPT . "?$1";
			$post->{'link'} =~ s/@{[POST_SCRIPT]}/@{[POST_SCRIPTNOC]}/ if ($opt_c);
		}
		# new scheme
		if ($post->{'link'} =~ m#$user/(\d+)\.html#) {
			$myhref = "$1.html";
			$post->{'link'} .= '?mode=reply' if ($opt_c);
		}
		
		if ($content = get_page($post->{'link'})) {
			$stat{'got_posts'}++;
			mkpath($dir, DEBUG_LEVEL, 0755);
			
			cleanup_html(\$content, $myhref, $user);
			
			rewrite_imgsrc(\$content, $up, $user) if ($opt_i);
			$content = from_utf8({-string => $content, -charset => LOCAL_CHARSET}) if (UTF8_DECODE || $opt_U);

			$content =~ s/<!-- keywords -->/<link rel="stylesheet" href="$up\/post.css" type="text\/css">\n<meta name="keywords" content="$post->{'keyword'}">\n$charset/;

			logmsg(">> $dir/$fname\n",2);
			open DF,">$dir/$fname" or die "error opening $dir/$fname for writing: $!\n";
			print DF $content;
			close DF;
				
			$post->{'status'} = 1;
			
		} else { # error fetching page
			last unless $opt_I;
		}
	}
}


=item rewrite_imgsrc(\$page, $up)

rewrite img's src attribute depending on $opt_i

=cut
sub rewrite_imgsrc {
	my ($page, $up, $user) = @_;
	my ($src, $extor, $d1, $d2);

	logmsg("extracting img src's...\n", 4);
	# extract all img src's
	$extor = new HTML::SimpleLinkExtor(BASE_URL);
	$extor->parse($$page);
	# list of unique image src's
	%images = (%images, map {$_ => 1} $extor->img());
	($d1, $d2) = split('\.', BASE_URL, 2);

	foreach $src (keys %images) {
		if ($src =~ m#stat\.${d2}img/(.*)$#) {
			if ($opt_i > 0) {
				$d1 = $1;
				$$page =~ s#src=["']\Q$src\E['"]#src='$up/img/$d1'#sg;
				$images{$src} = "img/$d1";
			} else {
				delete $images{$src};
			}
			
		} elsif ($src =~ m#userpic\.$d2(\d+/\d+)$#) {
			if ($opt_i > 1) {
				$d1 = $1;
				$$page =~ s#src=['"]\Q$src\E['"]#src='$up/userpic/$d1'#sg;
				$images{$src} = "userpic/$d1";
			} else {
				delete $images{$src};
			}
			
		} elsif ($opt_i > 2) {
			$d1 = $src;
			$d1 =~ s#^http://##;
			$d1 =~ s#[^(\w|\/|\.)]#_#g;
			$$page =~ s#src\s*=\s*['"]\Q$src\E['"]#src='$up/$user/img/$d1'#sg;
			$images{$src} = "$user/img/$d1";
			
		} else {
			delete $images{$src};
		}
	}	
}



sub cleanup_html {
	my ($page, $myhref, $user) = @_;
	my ($result, $in_navbar, $n, $in_reply, $extor, %links, $rlink);

	$in_navbar = 1;
	$in_reply = 0;
	$result = '';
	foreach (split(/\n/, $$page)) {
		# skip navigation bar
		if ($in_navbar && ($n = /<HR ?\/?>/i ... /<HR ?\/?>/i)) {
			$in_navbar = 0 if ($n =~ /E0/);
			next;
		}
		
		$in_reply = 0 if (m#</BODY>#i);
		next if $in_reply;
		$in_reply = 1 if ($opt_c && m#<a href=["']\Q$myhref\E['"]>Read comments</a>#);
		s#<H1>.*##i if $in_reply;
		next if /^<p>\[ <a href=.*<\/a>\]<\/p>$/;

		# add css link and keywords hook
		s#</HEAD>#<!-- keywords -->\n</HEAD>#i;

		# remove action buttons and forms
		# (delete, screen, mass action etc)
		s#</?form.*?>##g;
		s#<input .*?\#?>##g;
		s#<a href=['"]talkscreen.bml.*?</a>##g;
		s#<a href=['"]delcomment.bml.*?</a>##g;
		s#<label.*?</label>##g;
		s#<nobr>\s*</nobr>##g;
		s#<p><b>Mass action.*$##g;
		s#<link href=.*?>##g;
		
		$result .= "$_\n";
	}
	# replace relative hrefs with absolute
	$extor = new HTML::SimpleLinkExtor();
	$extor->parse($result);
	%links = map {$_ => 1} $extor->a();
	foreach (keys %links) {
		$rlink = $_;
		if ($rlink !~ m#^/#) {
			$rlink = "users/$user/$rlink";
		} else {
			$rlink =~ s#^/##;
		}
		$result =~ s#href=['"]\Q$_\E['"]#href='@{[BASE_URL]}$rlink'#sg if ($_ !~ m#^(http://|https://|ftp://)#);
	}
	$$page = $result;
	1;
}


=item lj_login()

POST login form to server, get cookies

=cut
sub lj_login {
	my($searchform, $loginform, $content);
	
	logmsg("logging in to " . BASE_URL . "... \n", 1);
	my ($user, $password) = ((defined $opt_u) && (length $opt_u > 0))? 
		split(":", $opt_u, 2) : (LOGIN, PASSWORD);
	
	# POST login form to get cookies
	if ($content = get_page(BASE_URL . LOGIN_SCRIPT)) {
		$loginform = parse HTML::Form($content, BASE_URL);
		# login form found
		if (defined $loginform && $loginform->find_input('user')) {
			$loginform->value('user',$user);
			$loginform->value('password',$password);

			#submit login form
			foreach (1 .. MAX_TRIES) {
				$res = $ua->request($loginform->click);

				if ($res->is_success) {
					$content = $res->content;
					# invalid login/password
					if ($content =~ /<span class="heading">Error<\/span>/) {
						logmsg("invalid login.\n", 0);
						return undef;
					}
					last;
					
				} else {
					my $err = $res->error_as_HTML;
					$err =~ s/^[^\d].*$//mg;
					$err =~ s/[\n\r]+//g;
					logmsg("\n$err\n", 0);
					logmsg("error submitting login form to " . $loginform->action ."\n", 0);
					logmsg("$_. retrying in 2 seconds...\n", 0);
					sleep 2;
				}
			}
			
		} else { # no login form 
			logmsg("got login page " . BASE_URL . LOGIN_SCRIPT . " but found no login form on it.\n", 0);
			return undef;
		}
		
		if ($res->is_success) {
			logmsg("got cookies.\n", 1);
			return 1;
		} else {
			return undef;
		}
		
	} else { # couldn't get login form page
		return undef;
	}
}


=item get_page($url)

download page from the remote host

=cut
sub get_page {
	my ($url, $is_image) = @_;

	if (!$is_image) {
		$url .= ($url =~ /\?/)? '&usescheme=lynx' : '?usescheme=lynx' 
			if ($url !~ /usescheme=lynx/);
	}
	logmsg("<< $url\n",2);
	$req = new HTTP::Request GET => $url;
	$req->header('Accept-Encoding' => 'gzip;q=1.0, *;q=0');
	
	foreach (1 .. MAX_TRIES) {
		logmsg("retrying $url...\n", 0) if ($_ > 1);
		#send request
		$res = $ua->request($req);
		#process responce
		if ($res->is_success) {
			$stat{'pages_ok'}++;
			
			return ($res->content_encoding && ($res->content_encoding =~ /gzip/))?
				Compress::Zlib::memGunzip($res->content) : $res->content;
			
		} else {
			my $err = $res->error_as_HTML;
			$err =~ s/^[^\d].*$//mg;
			$err =~ s/[\n\r]+//g;
			logmsg("\n$err\n$_. retrying in 2 seconds...\n", 0);
			sleep 2;
		}
	}
	$stat{'pages_err'}++;
	logmsg("failed to get $url after @{[MAX_TRIES]} attempts\n",0);
	# save failed downloads to log file
	print LF "Failed: $url\n";
	return undef;
}



sub logmsg {
	my ($message, $loglvl) = @_;
	if (!defined $loglvl) {
		print $message;
	} else {
		warn $message if ($loglvl <= DEBUG_LEVEL);
	}
}


=item build_index($user)

build index file for the given user

=cut
sub build_index {
	my ($user) = @_;
	my ($month, $year, @months);
	
	@months = ('','January','February','March','April','May','June',
	'July','August','September','October','November','December');

	# skip to next dir if there is no such user
	unless (-d LOCAL_DIR . $user) {
		logmsg(LOCAL_DIR . $user . "not found.");
		return;
	}
	
	# traverse directory tree calling process_html for each file found
	find({
			wanted => \&process_html_file,
			preprocess	=> \&sort_directory
		}, LOCAL_DIR . $user);

	# write index.html
	open DF, ">" . LOCAL_DIR . $user . "/index.html" 
		or die "error opening " . LOCAL_DIR . $user . "/index.html" .
				"for writing: $!\n";

	my $charset = (UTF8_DECODE || $opt_U)? LOCAL_CHARSET : 'utf-8';
	print DF <<EOH;
<html>
	<head>
		<title>Index file for $user livejournal</title>
		<meta http-equiv="Content-Type" content="text/html; charset=$charset">
		<link rel="stylesheet" href="index.css" type="text/css">
	</head>
	<body>
	<a name="top"></a>
	<hr width="550" size="3" noshade align="left">
	<font size="+2"><b><a href="@{[BASE_URL]}users/$user">$user</a></b>'s livejournal.&nbsp;&nbsp;</font>
EOH
	print DF "<font size=\"+1\">$stat{count_memos} <a href=\"#memories\">memories</a></font> "	
		if (scalar keys %memories);
		
	if (scalar keys %posts) {
		print DF "<font size=\"+1\"> | $stat{count_posts} posts: ";
		foreach (sort keys %posts) { # foreach year
			print DF "<a href=\"#$_\">$_</a> ";
		}
		print DF "</font>\n";
	}

	print DF "<br><tt><b>last updated:</b> " . (scalar localtime) . "</tt>\n";
	print DF '<hr width="550" size="3" noshade align="left">' . "\n";
	
	if (scalar keys %posts) {
		foreach $year (reverse sort keys %posts) { # $posts{$year} is a reference to the hash of months
			# year header
			print DF "<a name=\"$year\"></a>\n";
			print DF '<p><hr width="550" size="1" noshade align="left">' . "\n";
			print DF '<b><font size="+1"><a href="#top">' . $year . '</a>: </font></b><font size="-1">';
			print DF "<a href=\"#$year-$_\">" . 
				$months[$_+0] . "</a> | "
				foreach (sort {$a <=> $b} keys %{$posts{$year}});
			print DF '</font><hr width="550" size="1" noshade align="left"><br>' . "\n";

			# year body
			for  $month (reverse sort {$a <=> $b} keys %{$posts{$year}}) {
				print DF "<a name=\"$year-$month\"></a>[ <b>$months[$month]</b> ]<br>\n";
				print DF $posts{$year}->{$month} . "<p>\n";
			}
		}
	}

	if (scalar keys %memories) {
		print DF '<a name="memories"></a>'. "\n";
		print DF '<hr width="550" size="1" noshade align="left">' . "\n";
		print DF '<b><font size="+1"><a href="#top">Memories</a>: </font></b>' . "\n";
		print DF '<hr width="550" size="1" noshade align="left">' . "\n";
		print DF "<dl>\n<dt><b>$_</b></dt>\n" . $memories{$_} . "</dl>\n" 
			foreach (sort keys %memories);
	}
	
	print DF <<EOE;
<p><br>
<hr width="550" size="1" noshade align="left">
generated by <a href="http://www.offtopia.net/~ati/ljsm/">ljsm</a> @{[CVSVERSION]} 
</body>
</html>
EOE
	close DF or warn "Error closing index.html: $!\n";
}


# sort filenames so that the most recent posts go first
sub sort_directory {
	return sort {$b <=> $a} grep (/\w/, @_);
}

# callback subroutine for build_index
#
sub process_html_file {
	my ($line, $link, $kw, $title, $amuser, $itemid, $date, $locallink, $user);
	
	return unless ($File::Find::dir =~ m#(\w+)/(\d{4}/\d{1,2}|memories)#);
	$user = $1;
	return unless (-s && /\.html$/);
	
	
	# $_ is set to file name and we are inside target directory
	open DF, "<$_" or die "Error opening $File::Find::name for reading: $!\n";
	
	# search for link, keywords, title and date
	while ($line = <DF>) {
		$kw = $1 if ($line =~ /<meta name="keywords" content="(.*?)">/);
		$title = $1 if ($line =~ m#<font face=["']Arial,Helvetica['"] size=['"]?\+1['"]?><i><b>(.*?)</b></i>#i);
		$title = "<i>$1</i>" if ($line =~ m#<span class="heading">Error</span><br />(.*)$#i);
		$title = "<i>$1</i>" if ($line =~ m#^<H1>Error</H1><P>(.*)</P>$#i);
		$date = $1 if ($line =~ m{href="@{[BASE_URL]}users/\w+/day/\d\d\d\d/\d\d/(\d{1,2})"});
		$date = $1 if ($line =~ m{href="@{[BASE_URL]}users/\w+/\d\d\d\d/\d\d/(\d{1,2})/"});
		last if ($link);
	}
	
	$date = sprintf("%02d. ", $date) if $date;
	$link =~ s/&amp;/&/g if ($link); 
	close DF or warn "Error closing $File::Find::name : $!\n";
	$title = '<i>no title</i>' unless ($title);
	
	
	if ($File::Find::dir =~ /memories/) { # memories
		$stat{'count_memos'}++;

		$_ =~ m#(\w*)_(\d+)\.html#;
		($amuser,$itemid) = ($1, $2);
		$link = ($amuser)? "users/$amuser/$itemid.html?usescheme=lynx" : 
			"@{[POST_SCRIPT]}?itemid=$itemid&usescheme=lynx";

		$kw = '<i>default</i>' unless ($kw);
		
		# make separate link if there is a link it title (avoid nested <a>'s)
		$locallink = (index($title, '<a') > -1)? 
			"[<a href=\"memories/$_\" target=\"post\">read</a>]&nbsp; $title" :
			"<a href=\"memories/$_\" target=\"post\">$title</a>";
			
		$memories{$kw} .= "<dd><b><a href=\"@{[BASE_URL]}userinfo.bml?user=$amuser&mode=full&usescheme=lynx\">*</a>&nbsp;<a href=\"@{[BASE_URL]}users/$amuser/\">$amuser</a></b>: &nbsp; $locallink &nbsp;&nbsp;| <a href=\"@{[BASE_URL]}$link\"><b>&raquo;</b></a></dd>\n";
		
	} elsif ($File::Find::dir =~ m#(\w+)/(\d{4})/(\d{1,2})#)  { # posts 
		# $1 = user , $2 = year, $3 = month, $_ = html file name
		$stat{'count_posts'}++;
		$posts{$2} = {$3 => ''} if (!defined $posts{$2});

		# make separate link if there is a link in a title (avoid nested <a>'s)
		$locallink = (index($title, '<a') > -1)? 
			"[<a href=\"$2/$3/$_\" target=\"post\">read</a>]&nbsp; $title" :
			"<a href=\"$2/$3/$_\" target=\"post\">$title</a>";
			
		$posts{$2}->{$3} .= "<font color=\"gray\" size=\"-1\">$date</font> $locallink &nbsp;&nbsp;| <a href=\"@{[BASE_URL]}users/$1/$_?usescheme=lynx\"><b>&raquo;</b></a><br>\n";
		
	} else {
		# html file in unknown directory. just do nothing
	}
}

sub usage {
	warn <<EOW;
usage:
$0 [-r -m -a -O -I -u user:password -p proxyURL -d yyyy/mm[:yyyy/mm]] user1 user2 ...
$0 -x user1 user2 ...
$0 -X -u user:password [-r -O -I -p proxtURL -d yyyy/mm:yyyy/mm]
-a = save memories AND posts
-c = save posts without comments
-r = resume processing if there is already local file for the given post
-m = save memories instead of posts
-O = overwrite existing files (NOT recommended)
-i [1|2|3] = download icons (1) userpics (2) or all graphics (3) referenced in posts
-I = ignore network errors and continue fetching posts
-x = rebuild index file and exit
-X = export posts in XML format. Requires -u login:password
-u user:password = specify user:password pair for LJ login on the command prompt
-U = make UTF-8 to localt charset conversion
-p proxyURL = use proxyURL as a http proxy
-d yyyy/mm[:yyyy/mm] = save posts back to the specified date or in the specified date range
@{[CVSVERSION]}
EOW
}

# print some statistics and kiss goodbye
END {
	delete $stat{'count_posts'};
	delete $stat{'count_memos'};
	if ((DEBUG_LEVEL > 0) && (scalar keys %stat)) {
		print "\n\n================ s t a t i s t i c s ====================\n";
		print "ljsm.pl @{[CVSVERSION]}\n";
		print "$stat{$_} $_  " foreach keys %stat; 
		print "\n=========================================================\n";
	}
	close LF;
	#unlink 'ljsm.log' unless $stat{'pages_err'};
}
